/**
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * beck et al. projects GmbH
 * Theresienhoehe 13
 * 80336 Munich
 * Phone: +49-89-5442530
 * mailto: info@bea.de
 */
package de.bsvrz.ars.export;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

import de.bsvrz.ars.export.exceptions.InvalidObjectException;
import de.bsvrz.ars.export.prot.SqlProperties;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;

/**
 * Ein ExportProfil enthält alle Informationen, die für das Durchführen eines Exportes nötig sind.
 * Die Daten werden nach und nach gesetzt. Außerdem bietet die Klasse Methoden an, mit der die
 * Datenspezifikationen serialisiert/deserialisiert werden können.
 * 
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision: 1.1 $ / $Date: 2008/01/22 16:39:42 $ / ($Author: yvonnes $)
 */
public class ExportProfile {
    /**
     * Mögliche Formate
     */
    public enum Format {
        SQL, XML, LINE
    };


    /**
     * Mögliche Modi
     */
    public enum Modus {
        GUI, CMD
    };

    // ////////// Nutzdaten des ExportProfils
    private ClientDavInterface connection = null;
    private Modus modus;
    private DataspecificationList dataSpezificationList;
    private Format format = null;
    private long from = Long.MIN_VALUE;
    private String outputFile = null;
    private long until = Long.MIN_VALUE;
    private Collection AllObjects = new ArrayList();
    private SqlProperties sqlProperties = null;

    // //////// Datumsspezifische Angaben
    /**
     * Das Pattern für Datenangaben als String
     */
    public static final String datePattern = "TT.MM.JJJJJ SS:mm:ss";
    /**
     * Datumsformatierer
     */
    public static final DateFormat dateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");
    /**
     * Datumspattern. Mit/Ohne führenden Nullen, optionale Sekundenangabe
     */
    private static final Pattern regularDatePattern = Pattern
            .compile("(\\d{1,2})\\. *(\\d{1,2})\\. *(\\d{4,4}|\\d{2,2}) +(\\d{1,2}) *[\\.:] *(\\d{1,2}) *[\\.:]? *(\\d{1,2})?");
    private Matcher dateMatcher;

    /**
     * Singleton Instanz
     */
    public static final ExportProfile SINGLETON_INSTANCE = new ExportProfile();


    /**
     * Standard-Konstruktor für ein DatenExport. Private, da Singleton
     */
    private ExportProfile() {
        dateFormat.setLenient(false);
    }


    /**
     * Gibt dem ExportProfil ein Objekt, über das auf den Datenverteiler zugegriffen werden kann
     * 
     * @param connection DaV.
     */
    public void setDavConnection(ClientDavInterface connection) {
        dataSpezificationList = new DataspecificationList();
        this.connection = connection;
    }


    /**
     * liefert das Verbindungsobjekt zurück
     * 
     * @return Verbindungsobjekt zum DaV
     */
    public ClientDavInterface getDavConnection() {
        return connection;
    }


    /**
     * Liefert die Liste aller Objekte des DatenModels zurück (für die Auswahl)
     * 
     * @return Liste aller Objekte des DatenModels
     */
    public Collection getAllObjects() {
        return AllObjects;
    }


    /**
     * Die DatenspezifikationsListe (vom User ausgewählte Objekt, mit Namen)
     * 
     * @return DatenspezifikationsListe
     */
    public DataspecificationList getDataSpecificationList() {
        return dataSpezificationList;
    }


    /**
     * Liefert das Format als Enum (@link Format) zurück
     * 
     * @return das gewählte Format, ev null
     */
    public Format getFormat() {
        return format;
    }


    /**
     * Liefert den Anfangszeitpunkt als long-Wert
     * 
     * @return der Anfangszeitpunkt als long Wert, 0 wenn nicht gesetzt
     */
    public long getFrom() {
        return from;
    }


    /**
     * Liefert den Anfangszeitpunkt in textueller Form zum Darstellen für den User.
     * 
     * @return Anfangszeitpunkt, formatiert als String, oder leerer String, wenn nicht gesetzt
     */
    public String getFromAsString() {
        if (from > Long.MIN_VALUE) {
            return dateFormat.format(new Date(from));
        } else {
            return "";
        }
    }


    /**
     * Liefert ein File-Objekt für die Output-Datei
     * 
     * @return File-Objekt der Outputdatei
     */
    public String getOutputFile() {
        return outputFile;
    }

    
    public SqlProperties getSqlProperties() {
        return sqlProperties;
    }

    /**
     * Gibt an, ob der Export im GUI oder CMD Modus gestartet wurde
     * 
     * @return Modus
     */
    public Modus getModus() {
        return modus;
    }


    /**
     * Liefert den Endzeitpunkt als long-Wert
     * 
     * @return der Endzeitpunkt als long Wert, 0 wenn nicht gesetzt
     */
    public long getUntil() {
        return until;
    }


    /**
     * Liefert den Endzeitpunkt in textueller Form zum Darstellen für den User.
     * 
     * @return Endzeitpunkt, formatiert als String, oder leerer String, wenn nicht gesetzt
     */
    public String getUntilAsString() {
        if (until > Long.MIN_VALUE) {
            return dateFormat.format(new Date(until));
        } else {
            return "";
        }
    }


    /**
     * Prüft, ob ein Von-Zeitstempel gesetzt ist
     * 
     * @return Wahr, wenn ein Von-Zeitstempel gesetzt ist
     */
    public boolean hasFrom() {
        return (this.from == Long.MIN_VALUE);
    }


    /**
     * Prüft, ob ein Bis-Zeitstempel gesetzt ist
     * 
     * @return Wahr, wenn ein Bis-Zeitstempel gesetzt ist
     */
    public boolean hasUntil() {
        return (this.until == Long.MIN_VALUE);
    }


    /**
     * Lädt alle verfügbaren Objekte. Diese werden vorallem Benötigt, um die Listen der GUI zu
     * füllen
     * 
     * @return Wahr, wenn das Laden erfolgreich war
     */
    @SuppressWarnings("unchecked")
    public boolean loadAllObjects() {
        assert connection != null : "loadAllObjects, but no connection to DaV";

        if (AllObjects.size() == 0) {
            try {
                SystemObjectType sot = (SystemObjectType) connection.getDataModel().getObject("typ.typ");
                for (Iterator iter = sot.getObjects().iterator(); iter.hasNext();) {
                    SystemObjectType typ = (SystemObjectType) iter.next();
                    if (typ.isConfigurating() && typ.isBaseType()) {
                        AllObjects.addAll(typ.getObjects());
                    }
                }
                return true;
            } catch (ConfigurationException e) {
                AllObjects.clear();
            }
        }
        return false;
    }


    /**
     * Überprüft, ob dieses Exportprofil vollständig ist. Das heißt, es wird überprüft, ob alle
     * Felder gesetzt sind.
     * 
     * Es ist Aufgabe der Setter zu übeprüfen, dass die Daten in den Feldern semantisch korrekt
     * sind.
     * 
     * @return Liefert <code>true</code> zurück, wenn alle Informationen vorhanden sind, um einen
     * Export durchzuführen
     */
    public boolean isComplete() {
        if (format == null || from == Long.MIN_VALUE || outputFile == null || until == Long.MIN_VALUE) {
            return false;
        } else {
            return true;
        }
    }


    /**
     * File-Objekt für die XML-Datei, die die Datenspezifikation enthält. Diese Methode parst die
     * Datei und erzeugt entsprechende Java-Objekte.
     * 
     * @param dataSpezification Datei-Objekt
     * @throws FileNotFoundException Wenn die Datei nicht gefunden wurde
     * @throws SAXException Jede Art von Perser Exception
     * @throws IOException Bei Zugriffsproblemen
     * @return List an Strings mit Fehlermeldung. Wenn die Liste leer ist, gab es keine Fehler
     */
    public List<String> loadDataspecificationFile(File dataSpezification) throws FileNotFoundException, SAXException, IOException {
        assert connection != null : "loadDataspecificationFile, but no connection to DaV";
        try {
            XMLReader xr = XMLReaderFactory.createXMLReader();
            XML2Dataspezification x2d = new XML2Dataspezification(connection);
            xr.setContentHandler(x2d);
            xr.setErrorHandler(x2d);
            FileReader r = new FileReader(dataSpezification);
            InputSource is = new InputSource(r);
            xr.parse(is);
            if (x2d.hasErrors()) {
                System.out.println("Fehler: " + x2d.getErrors());
                dataSpezificationList = new DataspecificationList();
                return x2d.getErrors();
            } else {
                dataSpezificationList.replaceDataspecificationWith(x2d.getDataspecificationList());
                return new ArrayList<String>();
            }
        } catch (FileNotFoundException e) {
            throw e;
        } catch (SAXException e) {
            throw e;
        } catch (IOException e) {
            throw e;
        } catch (Exception e) {
            throw new IOException();
        }
    }


    /**
     * File-Objekt für die XML-Datei, die die Datenspezifikation enthält.
     * 
     * @see ExportProfile#loadDataspecificationFile(File dataSpezification)
     * @param dataSpezification Dateiname als String
     * @throws FileNotFoundException Wenn die Datei nicht gefunden wurde
     * @throws SAXException Jede Art von Perser Exception
     * @throws IOException Bei Zugriffsproblemen
     * @return List an Strings mit Fehlermeldung. Wenn die Liste leer ist, gab es keine Fehler
     */
    public List<String> loadDataspecificationFile(String dataSpezification) throws FileNotFoundException, SAXException, IOException {
        return loadDataspecificationFile(new File(dataSpezification));
    }


    /**
     * Setzt das Format (XML bzw. SQL) mit einem Wert aus
     * 
     * @param format Das Format
     */
    public void setFormat(Format format) {
        this.format = format;
    }


    /**
     * Setzt das Format (XML bzw. SQL) für den Export Der String ist Caseinsensitive.
     * 
     * @param format Format.
     */
    public void setFormat(String format) {
        if (format.equalsIgnoreCase("xml")) {
            this.format = Format.XML;
        } else if (format.equalsIgnoreCase("sql")) {
            this.format = Format.SQL;
        } else if (format.equalsIgnoreCase("line")) {
            this.format = Format.LINE;
        } else {
            // FIXME Exception
        }
    }


    /**
     * Setzt die Anfangszeit
     * 
     * @param from Anfangszeit als long-Wert
     */
    public void setFrom(long from) {
        this.from = from;
    }


    /**
     * Setzt die Anfangszeit
     * 
     * @param from Anfangszeit als String im Format TT.MM.JJJJ SS:MM:ss
     * @throws ParseException Exception, wenn der String nicht in ein Datum umgewandelt werden kann
     */
    public void setFrom(String from) throws ParseException {
        try {
            this.from = dateFormat.parse(parseDateString(from)).getTime();
        } catch (ParseException e) {
            this.from = 0L;
            throw e;
        }
    }


    /**
     * Setzt die Ausgabedatei
     * 
     * @param outputFile Ausgabedatei
     */
    public void setOutputFile(File outputFile) {
        this.outputFile = outputFile.toString();
    }


    /**
     * Setzt die Ausgabedatei
     * 
     * @param outputFile Ausgabedatei
     */
    public void setOutputFile(String outputFile) {
        this.outputFile = outputFile;
    }


    /**
     * Setz den Modus (CUI/CMD)
     * 
     * @param modus Modus (CUI/CMD)
     */
    public void setModus(Modus modus) {
        this.modus = modus;
    }

    
    /**
     * SQL-Properties, wenn nicht der Default genommen werden soll
     *
     * @param p SqlProperties.
     */
    public void setSqlProperties(SqlProperties p) {
        sqlProperties = p;
    }

    /**
     * Setzt den Endzeitpunkt
     * 
     * @param until Endzeitpunkt als long-Wert
     */
    public void setUntil(long until) {
        this.until = until;
    }


    /**
     * Setzt den Endzeitpunkt
     * 
     * @param until Endzeitpunkt als String im Format TT.MM.JJJJ SS:MM:ss
     * @throws ParseException Exception, wenn der String nicht als Datum interpretiert werden kann
     */
    public void setUntil(String until) throws ParseException {
        try {
            this.until = dateFormat.parse(parseDateString(until)).getTime();
        } catch (ParseException e) {
            this.until = Long.MIN_VALUE;
            throw e;
        }
    }


    /**
     * Textuelle Repräsentation dieses Objekts
     * 
     * @see java.lang.Object#toString()
     * @return String mit den wichtigsten Informationen, die in diesem Objekt gespeichert sind
     */
    public String toString() {
        StringBuffer s = new StringBuffer(256);
        s.append("[");
        s.append("datenspez=" + ((dataSpezificationList == null) ? "null" : dataSpezificationList.getSize() + " Elemente")); // FIXME
        // namen
        // laden
        s.append("; ");
        s.append("ausgabe=" + ((outputFile == null) ? "null" : outputFile));
        s.append("; ");
        s.append("format=" + ((format == null) ? "null" : format.toString().toLowerCase()));
        s.append("; ");
        s.append("von=" + ((from == Long.MIN_VALUE) ? "null" : dateFormat.format(from)));
        s.append("; ");
        s.append("bis=" + ((until == Long.MIN_VALUE) ? "null" : dateFormat.format(until)));
        s.append("]");

        return s.toString();
    }


    /**
     * Gibt an, ob die GUI für die Eingabe weiterer Daten benötigt wird.
     * 
     * @return Ob das GUI geöffnet werden muss oder nicht 
     */
    public boolean openGui() {
        assert modus != null : "Kein Modus gesetzt!";
        return (modus == Modus.GUI);
    }


    /**
     * Parst einen String als Datum, dabei werden auch unterschiedliche Punktionszeichen unterstützt
     * 
     * @param s String, der ein Datum darstellt
     * @return String, der von den Parsemethoden der Sun Klassen interpretiert werden kann
     * @throws ParseException wenn der String nicht als Datum geparst werden kann
     */
    private String parseDateString(String s) throws ParseException {
        StringBuffer sb = new StringBuffer(21);
        dateMatcher = regularDatePattern.matcher(s);
        if (dateMatcher.matches()) {
            // Tag
            sb.append(dateMatcher.group(1));
            sb.append(".");
            // Monat
            sb.append(dateMatcher.group(2));
            sb.append(".");
            // Jahr
            if (dateMatcher.group(3).length() == 2) {
                sb.append("20"); // Zweistellig ist 20xx
            }
            sb.append(dateMatcher.group(3));
            sb.append(" ");
            // Stunde
            sb.append(dateMatcher.group(4));
            sb.append(":");
            // Minute
            sb.append(dateMatcher.group(5));
            sb.append(":");
            // Sekunde
            if (dateMatcher.group(6) == null) {
                sb.append("00");
            } else {
                sb.append(dateMatcher.group(6));
            }
            return sb.toString();

        } else {
            throw new ParseException("Ungültiges Datumsformat", 0);
        }
    }


    /**
     * Innere Klasse, die aus dem XML Dokument wieder Datenspezifikationen erstellt
     * 
     * @author beck et al. projects GmbH
     * @author Phil Schrettenbrunner
     * 
     */
    private class XML2Dataspezification extends DefaultHandler {
        private ClientDavInterface con;
        private ArrayList<SystemObject> sysObject = new ArrayList<SystemObject>();

        private List<Dataspecification> dsl;
        private String DsName;
        private DataDescription dd;

        private List<String> errors;


        /**
         * Konstruktur
         * 
         * @param connection DaV Verbindung
         */
        public XML2Dataspezification(ClientDavInterface connection) {
            dsl = new ArrayList<Dataspecification>();
            con = connection;
            errors = new ArrayList<String>();
        }


        /**
         * Gibt an, ob es beim Einlesen zu Schwierigkeiten gekommen ist und ob die Daten somit ev.
         * nicht gültig sind.
         * 
         * @return true, wenn es zu Fehlern gekommen ist
         */
        public boolean hasErrors() {
            return (errors.size() > 0);
        }


        /**
         * Das Array mit den Fehlermeldungen
         * 
         * @return Liste der aufgetretenen Fehler
         */
        public List<String> getErrors() {
            return errors;
        }


        /**
         * Reagert auf die Start-Tags und beginnt, Daten zu sammeln
         * 
         * @param uri The Namespace URI, or the empty string if the element has no Namespace URI or
         * if Namespace processing is not being performed.
         * @param localName The local name (without prefix), or the empty string if Namespace
         * processing is not being performed.
         * @param qName The qualified name (with prefix), or the empty string if qualified names are
         * not available.
         * @param attributes The attributes attached to the element. If there are no attributes, it
         * shall be an empty Attributes object.
         * @exception org.xml.sax.SAXException Any SAX exception, possibly wrapping another
         * exception.
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
            if (qName.equals("datenspezifikation")) {
                // Name
                DsName = attributes.getValue("name");
                // AttributGruppe

                long asp = 0L, atg = 0L;
                short simvar = -1;

                // Aspekt
                try {
                    asp = Long.parseLong(attributes.getValue("asp"));
                } catch (NumberFormatException e) {
                    errors.add("'" + attributes.getValue("asp") + "' ist keine Aspekt-ID bei Datenspezifikation '" + DsName + "'");
                }

                // AttributGruppe
                try {
                    atg = Long.parseLong(attributes.getValue("atg"));
                } catch (NumberFormatException e1) {
                    errors.add("'" + attributes.getValue("atg") + "' ist keine Attributgruppen-ID bei Datenspezifikation '" + DsName + "'");
                }

                // SimVar
                try {
                    simvar = Short.parseShort(attributes.getValue("simvar"));
                } catch (NumberFormatException e) {
                    errors.add("Ungültige SimulationsVariante '" + attributes.getValue("simvar") + "' bei Datenspezifikation '" + DsName + "'");
                }

                // DD erzeugen
                try {
                    AttributeGroup atgInstance = (AttributeGroup) con.getDataModel().getObject(atg);
                    Aspect aspInstance = (Aspect) con.getDataModel().getObject(asp);

                    if (atgInstance == null) {
                        errors.add("Die AttributGruppe '" + atg + "' der Datenspezifikation '" + DsName + "' existiert nicht.");
                    }
                    if (aspInstance == null) {
                        errors.add("Der Aspekt '" + asp + "' der Datenspezifikation '" + DsName + "' existiert nicht.");
                    }

                    dd = new DataDescription(atgInstance, aspInstance, simvar);
                } catch (ConfigurationException e) {
                    dd = null;
                    errors.add("Fehler beim Erzeugen der Datenbeschreibung bei Datenspezifikation '" + DsName + "'");
                }
            } else if (qName.equals("objekt")) {
                // Ein Objekt unterhalb einer Datenspezifikation
                long l = Long.parseLong(attributes.getValue("id"));
                try {
                    SystemObject so = (SystemObject) con.getDataModel().getObject(l);
                    if (so == null) {
                        errors.add("SystemObjekt mit id '" + l + "' kann nicht geladen werden");
                    } else {
                        sysObject.add(so);
                    }
                } catch (ConfigurationException e) {
                    errors.add("Konfigurations-Ausnahme '" + e.getMessage() + "'");
                } catch (NullPointerException e) {
                    errors.add("NullPointer Ausnahme");
                }
            }
        }


        /**
         * Reagiert auf die EndTags und erstellt aus den Daten, die startElement(String, String,
         * String, Attributes) sammelt, Java Objekte.
         * 
         * @see org.xml.sax.ContentHandler#endElement(java.lang.String, java.lang.String,
         * java.lang.String)
         */
        public void endElement(String uri, String localName, String qName) throws SAXException {
            if (qName.equals("datenspezifikation")) {
                try {
                    dsl.add(new Dataspecification(sysObject, dd, DsName));
                } catch (NullPointerException e) {
                    errors.add("NullPointer Ausnahme");
                    throw e;
                } catch (InvalidObjectException e) {
                    errors.add("Ungültiges Objekt: " + e.getMessage());
                }

                // Verwendete Objekte zurücksetzen
                DsName = "";

                sysObject = new ArrayList<SystemObject>();
            }
        }


        /**
         * Liefert die Liste der aus der Datei gelesenen Datenspezifikationen
         * 
         * @return Liste der Datenspezifikationen
         */
        public List<Dataspecification> getDataspecificationList() {
            return dsl;
        }
    }
}
